#!/usr/bin/python

import math

from .. import utils


def _unity_gain_propagation_loss_db(model):
    # Sklar, page 254, Equation 5.10
    d = model.slant_range_km * 1000
    lam = model.wavelength_m
    return 2 * utils.to_db(4 * math.pi * d / lam)


def _allocation_start_hz(model):
    return model.center_freq_hz - (model.allocation_hz/2)


def _allocation_end_hz(model):
    return model.center_freq_hz + (model.allocation_hz/2)


def _total_channel_loss_db(model):
    return (model.unity_gain_propagation_loss_db
            + model.atmospheric_loss_db
            + model.ionospheric_loss_db
            + model.rain_loss_db
            + model.multipath_fading_db)


def _required_tx_bw_hz(model):
    return model.bitrate_hz / model.tx_spectral_efficiency_bps_per_hz


def _required_rx_bw_hz(model):
    return model.bitrate_hz / model.rx_spectral_efficiency_bps_per_hz


def _required_tx_bw_dbhz(model):
    return utils.to_db(model.required_tx_bw_hz)


def _required_rx_bw_dbhz(model):
    return utils.to_db(model.required_rx_bw_hz)


def _wavelength_m(model):
    return model.speed_of_light_m_per_s / model.center_freq_hz


def _bitrate_dbhz(model):
    return utils.to_db(model.bitrate_hz)


class Channel(object):

    def __init__(self,
                 center_freq_mhz=402.7,
                 bitrate_hz=9600,
                 atmospheric_loss_db=0.5,
                 ionospheric_loss_db=0.5,
                 rain_loss_db=0.5,
                 multipath_fading_db=0,
                 allocation_hz=10e6,
                 speed_of_light_m_per_s=299792458.00,
                 polarization_mismatch_loss_db=3.0,
                 gs_pfd_limits=None):

        # FIXME: Consider calculating polarization mismatch loss from
        # angle, including axial ratio, ...

        self.tribute = {
            # calculators
            'unity_gain_propagation_loss_db': _unity_gain_propagation_loss_db,
            'total_channel_loss_db': _total_channel_loss_db,
            'wavelength_m': _wavelength_m,
            'bitrate_dbhz': _bitrate_dbhz,
            'required_tx_bw_hz': _required_tx_bw_hz,
            'required_rx_bw_hz': _required_rx_bw_hz,
            'required_tx_bw_dbhz': _required_tx_bw_dbhz,
            'required_rx_bw_dbhz': _required_rx_bw_dbhz,
            'allocation_start_hz': _allocation_start_hz,
            'allocation_end_hz': _allocation_end_hz,
            
            # constants
            'center_freq_hz': center_freq_mhz * 1e6,
            'bitrate_hz': bitrate_hz,
            'atmospheric_loss_db': atmospheric_loss_db,
            'ionospheric_loss_db': ionospheric_loss_db,
            'rain_loss_db': rain_loss_db,
            'multipath_fading_db': multipath_fading_db,
            'speed_of_light_m_per_s': speed_of_light_m_per_s,
            'polarization_mismatch_loss_db': polarization_mismatch_loss_db,
            'allocation_hz': allocation_hz,
            'gs_pfd_limits': gs_pfd_limits,
            }



#!/usr/bin/python

import math

from ..model import DAGModel
from .. import utils


class Interconnect(object):
    """Provides the specification for the RF chain between two components.

    For example, you might include cabling, connectors, cavity
    filters, and LNAs between a ground-station antenna and a modem.
    You might also create one that specifies the cable between a
    space-side radio module, and space-side antenna module.  Doing it
    this way allows us to specify the rf components on individual
    components, and connections between those components separately.

    Whether or not this interconnect is intended for use on the rx
    side is needed so it can be incorporated in the rx cascade
    calculations or included in the tx inline losses after the PA.

    The overall model assumed here is one of:

    rx_antenna -> rx_interconnect -> receiver

    transmitter -> tx_interconnect -> tx_antenna

    Doing so this way means we have exactly one interconnect per
    budgeted link.
    """

    def __init__(self,
                 name=None,
                 rf_chain=[],
                 is_rx=True):

        # preserve the input
        self.is_rx = is_rx
        self.rf_chain = rf_chain

        if not name:
            name = 'RX Interconnect' if is_rx else 'TX Interconnect'

        self.name = name

        self.tribute = {
            # calculators

            # constants
            self._name('rf_chain'): rf_chain,
            self._name('name'): name,
            }

    def _name(self, s):
        if self.is_rx:
            return 'rx_interconnect_'+s
        else:
            return 'tx_interconnect_'+s



#!/usr/bin/env python

import pylink
import pytest


@pytest.fixture
def example_rf_chain():
    return [
        pylink.Element(gain_db=-0.1,
                       noise_figure_db=0.1,
                       name='attenuator')
        ]


@pytest.fixture
def attenuator_1db():
    return pylink.Element(gain_db=-1,
                          noise_figure_db=1,
                          name='attenuator')

@pytest.fixture
def amplifier_10db():
    return pylink.Element(gain_db=10,
                          noise_figure_db=3,
                          name='amplifier')


perf = [
    pylink.Code("BPSK", .5, .5, 4),
    pylink.Code("QPSK", 1, 1, 8),
    pylink.Code("8PSK", 2, 2, 13),
    ]


@pytest.fixture
def model():
    return pylink.DAGModel([pylink.Geometry(),
                            pylink.Antenna(is_rx=True, tracking=True),
                            pylink.Interconnect(is_rx=True),
                            pylink.Receiver(),
                            pylink.Transmitter(),
                            pylink.Interconnect(is_rx=False),
                            pylink.Antenna(is_rx=False, tracking=False),
                            pylink.Channel(),
                            pylink.Modulation(name='QPSK', perf=perf),
                            pylink.LinkBudget()])



#!/usr/bin/python

"""Link Budget Management in Python

This package provides a flexible system with which to build and solve
any Directed Acyclic Graph data model.  The design objective is to
hide the complexity of calculation in discrete segments all registered
through a simple central interface.  See the package README for more
details, the examples directory in the source code for how to use it.

IGNORE THE 'PACKAGE CONTENTS'.  Python is annoying, here's what you
can look up if you're interested:

=== Tributaries ===
Antenna
LinkBudget
Channel
Geometry
Interconnect
Code
Modulation
Receiver
Transmitter

=== Other Objects ===
Element
DAGModel
LoopException
BitrateFigure
CanonicalPFDFigure
ExpectedPFDFigure
PFDvsBWFigure
BitrateFigure
Report
TaggedAttribute

=== Utility Functions ===
to_db
from_db
spreading_loss_db
pattern_generator
rx_pfd_hz_adjust
tx_pfd_hz_adjust
e_field_to_eirp_dbw
eirp_dbw_to_e_field_v_per_m
human_hz
human_m
"""

__title__ = 'pylink'
__author__ = 'Harrison Caudill <harrison@hypersphere.org>'
__version__ = '0.6'
__license__ = 'BSD'


from pylink.element import Element
from pylink.model import DAGModel
from pylink.model import LoopException
from pylink.report import BitrateFigure
from pylink.report import CanonicalPFDFigure
from pylink.report import ExpectedPFDFigure
from pylink.report import PFDvsBWFigure
from pylink.report import BitrateFigure
from pylink.report import Report
from pylink.tagged_attribute import TaggedAttribute
from pylink.utils import to_db
from pylink.utils import from_db
from pylink.utils import spreading_loss_db
from pylink.utils import pattern_generator
from pylink.utils import rx_pfd_hz_adjust
from pylink.utils import tx_pfd_hz_adjust
from pylink.utils import e_field_to_eirp_dbw
from pylink.utils import eirp_dbw_to_e_field_v_per_m
from pylink.utils import human_hz
from pylink.utils import human_m
from pylink.utils import human_b

from pylink.tributaries.antenna import Antenna
from pylink.tributaries.budget import LinkBudget
from pylink.tributaries.channel import Channel
from pylink.tributaries.geometry import Geometry
from pylink.tributaries.interconnect import Interconnect
from pylink.tributaries.modulation import Code
from pylink.tributaries.modulation import Modulation
from pylink.tributaries.modulation import NORMAL_DVBS2X_PERFORMANCE
from pylink.tributaries.modulation import PERFECT_DVBS2X_PERFORMANCE
from pylink.tributaries.receiver import Receiver
from pylink.tributaries.transmitter import Transmitter
from pylink.tributaries.hyperspectral import HyperSpectralSNRBudget



#!/usr/bin/python

import scipy
import scipy.interpolate
import scipy.signal
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import math

from ..model import DAGModel
from .. import utils


def _floor(v, n):
    return int(n * math.floor(v/n))


def _find_nearest_index(array, value):
    value = (360 + value) % 360
    return np.abs(array - value).argmin()


def _average_gain_dbi(pattern, angles):
    return sum(pattern) / float(len(pattern))


def _average_nadir_gain_dbi(pattern, angles):
    """Average gain on the nadir face of the satellite.

    For simplicity, this function assumes some hard-coded values of
    65-degrees off of boresight.  That translates to 0->65 and (360-65)->360
    """
    s = 0
    n = 0
    offset = 65
    for i in range(len(pattern)):
        angle = angles[i]
        gain = pattern[i]
        if (0 <= angle <= offset) or ((360-offset) <= angle <= 360):
            s += gain
            n += 1
    return s / n


class Antenna(object):
    """Antenna tributary

    This class can be used either for tx or for rx and it will
    register its functions as either the tx_antenna_... or
    rx_antenna_... as appropriate.
    """

    def __init__(self,
                 pattern=None,
                 gain=0.0,
                 polarization='RHCP',
                 tracking=True,
                 rf_chain=[],
                 pointing_loss_db=0,
                 is_rx=True,
                 **meta):
        """Create a new antenna tributary.

        pattern -- list of evenly-spaced pattern cut values starting at 0
        gain -- peak gain of the antenna
        polarization -- str
        tracking -- does it track the target (eg rotator) or not (eg nadir)
        rf_chain -- list of Element objects for the RF hain on the board
        pointing_loss_db -- for now, just the number of dB of pointing loss
        is_rx -- is it for receive or transmit
        kwargs -- any metadata to assign to the antenna itself

        If there are 360 points in the pattern, it will be
        interpolated for you automatically.
        """

        self.meta = meta

        self.peak_gain_only = (pattern is None)
        if pattern is None:
            self.peak_gain_only = True
            self.peak_gain = gain
            pattern = np.zeros(360)
            pattern += gain
        else:
            self.peak_gain = max(pattern)

        pattern = np.array(pattern)

        self.pattern_angles = np.arange(0.0, 360.0, 360.0/len(pattern))
        self.pattern = pattern

        if len(pattern) == 360:
            self.interpolated = pattern[:]
            self.interpolated_angles = np.arange(0, 360, 1)
        else:
            interpolated = self._interpolate_pattern(pattern)
            self.interpolated_angles = np.arange(0, 360, 360/len(interpolated))
            self.interpolated = interpolated

        self.is_rx = is_rx

        self.tribute = {
            # calculators
            self._mangle('peak_gain_dbi'): self._peak_gain_dbi,
            self._mangle('gain_dbi'): self._gain_dbi,
            self._mangle('angle_deg'): self._angle_deg,
            self._mangle('boresight_gain_dbi'): self._boresight_gain_dbi,
            self._mangle('average_gain_dbi'): self._average_gain_dbi,
            self._mangle('average_nadir_gain_dbi'): self._average_nadir_gain_dbi,

            # constants
            self._name('polarization'): polarization,
            self._name('raw_gain_pattern'): pattern,
            self._name('raw_gain_pattern_angles'): self.pattern_angles,
            self._name('gain_pattern'): self.interpolated,
            self._name('gain_pattern_angles'): self.interpolated_angles,
            self._name('obj'): self,
            self._name('tracking_target'): not not tracking,
            self._name('rf_chain'): rf_chain,
            self._name('pointing_loss_db'): pointing_loss_db,
            }

    def _name(self, s):
        if self.is_rx:
            return 'rx_antenna_'+s
        else:
            return 'tx_antenna_'+s

    def _lst_to_rad(self, lst):
        return np.array([math.radians(v) for v in lst])

    def _wrap(self, lst):
        return np.array(list(lst) + [lst[0]])

    def _plot_peak_gain(self, fname, title):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection='polar')

        theta = self._lst_to_rad(self.pattern_angles[:])
        pattern = np.array(self.pattern)

        # offset the pattern to get around the negative-radius issue
        if self.peak_gain < 0:
            offset = -2 * self.peak_gain
            pattern += offset

        ax.plot(theta,
                pattern,
                color='r',
                linewidth=3,
                label='Peak Gain Used Everywhere')
        fig.canvas.draw()
        if self.peak_gain < 0:
            ax.set_yticklabels([t - offset for t in ax.get_yticks()])

        fig.suptitle(title)
        plt.legend(loc=4)

        fig.savefig(fname, transparent=True)

    def _plot_interpolated(self, fname, title, include_raw, ylim):
        # Wrap around one point to close the loop and convert to radians
        interp = self._wrap(self.interpolated)
        raw = np.copy(self.pattern)

        low = min(min(interp), min(raw))
        hi = max(min(interp), max(raw))

        n_steps = 5
        min_step_size = 1
        step_size = max(int((hi - low) / n_steps), min_step_size)

        low_r = _floor(low, step_size)
        hi_r = _floor(hi, step_size)

        val_start = low_r if low_r < low else low_r - step_size
        val_stop = hi_r + step_size

        offset = 0 - val_start

        # to debug uncomment these lines
        # print 'low:         %s' % low
        # print 'hi:          %s' % hi
        # print 'low_r:       %s' % low_r
        # print 'hi_r:        %s' % hi_r
        # print 'val_start:   %s' % val_start
        # print 'val_stop:    %s' % val_stop
        # print 'step_size:   %s' % step_size
        # print 'offset:      %s' % offset
        # print

        interp += offset
        raw += offset

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection='polar')

        if ylim:
            locator = matplotlib.ticker.MaxNLocator(nbins=8)
            ax.yaxis.set_major_locator(locator)            
            ax.set_ylim([ylim[0]+offset, ylim[1]+offset])

        interp_angles = self._wrap(self._lst_to_rad(self.interpolated_angles))
        raw_angles = self._lst_to_rad(self.pattern_angles)

        include_raw = (include_raw
                       and (len(self.pattern) != len(self.interpolated)))

        if len(self.pattern) == len(self.interpolated):
            label = 'Antenna Pattern'
            main_angles = raw_angles
            main_pattern = raw
        else:
            label = 'Interpolated Pattern'
            main_angles = interp_angles
            main_pattern = interp


        ax.set_theta_zero_location("N")

        ax.plot(main_angles,
                main_pattern,
                color='r',
                linewidth=3,
                label=label)

        if include_raw:
            ax.plot(raw_angles,
                    raw, 'x',
                    color='b',
                    linewidth=1,
                    label='Observed')
        fig.canvas.draw()

        ax.set_yticklabels([t - offset for t in ax.get_yticks()])

        fig.suptitle(title)
        plt.legend(loc=4)

        fig.savefig(fname, transparent=True)

    def plot_pattern(self, fname, include_raw=True, title=None, ylim=None):
        """Plots the pattern to a PNG file.

        fname -- where to save it
        include_raw -- If the pattern is interpolated, include the raw points?
        title -- Title of the image
        ylim -- [min, max] as desired

        If, for example, your real pattern varies by only one dB, its
        plot can be correct, but look a little weird as you see it
        vary wildly from one side to the other whereas it is quite
        stable in reality.  Thtat's why the <ylim> is an option.
        """

        prefix = 'RX' if self.is_rx else 'TX'
        if not title:
            title = '%s Antenna Gain Pattern' % prefix

        if self.peak_gain_only:
            return self._plot_peak_gain(fname, title)
        else:
            return self._plot_interpolated(fname, title, include_raw, ylim)

    def _linear_interpolate(self, src, factor):
        src_x = np.arange(0, len(src), 1)
        tck = scipy.interpolate.splrep(src_x, src, s=0)
        dst_x = np.arange(0, len(src), 1.0/factor)
        dst = scipy.interpolate.splev(dst_x, tck, der=0)
        return dst

    def _circular_interpolate(self, src, factor):
        tmp = list(src)*3
        tmp = self._linear_interpolate(tmp, factor)
        l = int(len(tmp) / 3)
        return tmp[l:2*l]

    def _interpolate_pattern(self, pattern, factor=None):
        if not factor:
            # default to roughly every one degree
            factor = (360.0 / len(pattern))
        return self._circular_interpolate(pattern, factor)

    def _mangle(self, name):
        x = 'rx' if self.is_rx else 'tx'
        s = '_' if name[0] == '_' else ''
        return '%s%s_antenna_%s' % (s, x, name)

    def _call(self, model, name):
        return getattr(model, self._mangle(name))

    def _peak_gain_dbi(self, model):
        return max(self._call(model, 'gain_pattern'))

    def _gain_dbi(self, model):
        if self._call(model, 'tracking_target'):
            return self._call(model, 'boresight_gain_dbi')
        else:
            angle = self._call(model, 'angle_deg')
            angles = self._call(model, 'gain_pattern_angles')
            idx = _find_nearest_index(angles, angle)
            pattern = self._call(model, 'gain_pattern')
            return pattern[idx]

    def _angle_deg(self, model):
        if self._call(model, 'tracking_target'):
            return 0

        if model.is_downlink:
            if self.is_rx:
                # We are the ground-station
                return model.min_elevation_deg
            else:
                # We are the satellite
                return model.satellite_antenna_angle_deg
        else:
            if self.is_rx:
                # We are the satellite
                return model.satellite_antenna_angle_deg
            else:
                # We are the ground-station
                return model.min_elevation_deg

    def _boresight_gain_dbi(self, model):
        pattern = self._call(model, 'gain_pattern')
        angles = self._call(model, 'gain_pattern_angles')
        idx = _find_nearest_index(angles, 0)
        return pattern[idx]

    def _average_gain_dbi(self, model):
        pattern = self._call(model, 'gain_pattern')
        angles = self._call(model, 'gain_pattern_angles')
        return _average_gain_dbi(pattern, angles)

    def _average_nadir_gain_dbi(self, model):
        pattern = self._call(model, 'gain_pattern')
        angles = self._call(model, 'gain_pattern_angles')
        return _average_nadir_gain_dbi(pattern, angles)



#!/usr/bin/python

import math

from .. import utils


def _tx_eirp_dbw(model):
    return (model.tx_power_at_antenna_dbw
            + model.tx_antenna_gain_dbi
            - model.tx_antenna_pointing_loss_db)


def _peak_tx_eirp_dbw(model):
    return model.tx_power_at_antenna_dbw + model.tx_antenna_peak_gain_dbi


def _tx_power_at_antenna_dbw(model):
    return model.tx_power_at_pa_dbw - model.tx_inline_losses_db


def _tx_inline_losses_db(model):
    chain = (model.tx_antenna_rf_chain
             + model.tx_interconnect_rf_chain
             + model.transmitter_rf_chain)
    return -1 * sum([c.gain_db for c in chain])


def _rx_rf_chain(model):
    return (model.rx_antenna_rf_chain
            + model.rx_interconnect_rf_chain
            + model.receiver_rf_chain)


def _pf_dbw_per_m2(model):
    spreading = utils.spreading_loss_db(model.slant_range_km)
    return (model.tx_eirp_dbw
            - spreading
            - model.atmospheric_loss_db
            - model.ionospheric_loss_db
            - model.rain_loss_db)


def _pfd_dbw_per_m2_per_hz(model):
    return utils.rx_pfd_hz_adjust(model, model.pf_dbw_per_m2, 1)


def _canonical_pf_dbw_per_m2(model):
    # this one is here for compliance.  We want to show that under
    # maximum power flux conditions, we can meet our pfd limitations.
    # We ignore things like atmospheric loss and antenna pattern, but
    # we do count slant range.
    spreading = utils.spreading_loss_db(model.slant_range_km)
    return (model.peak_tx_eirp_dbw - spreading)


def _peak_pf_dbw_per_m2(model):
    # no matter the tx/rx configuration, the peak pfd is always when
    # the satellite is overhead at the periapsis
    spreading = utils.spreading_loss_db(model.periapsis_altitude_km)

    # The peak EIRP is what we shall consider
    eirp = model.peak_tx_eirp_dbw

    # Regarding occupied bandwidth: The PFD graph assumes that we
    # fully occupy our allocation.  For worst-case, peak PFD purposes,
    # we're going to assume that the currently occupied BW (which
    # should probably be set to the lowest bitrate we'd ever use), is
    # what we'll use.  As such, there is no need to adjust bitrate or
    # anything else.

    # To be conservative, we'll ignore *spheric and rain loss
    return (eirp - spreading)


def _peak_pfd_dbw_per_m2_per_hz(model):
    return utils.pfd_hz_manual_adjust(model.peak_pf_dbw_per_m2,
                                      model.allocation_hz,
                                      1)


def _range_to_geo_km(model):
    return model.geo_altitude_km - model.apoapsis_altitude_km


def _peak_pf_at_geo_dbw_per_m2(model):
    if model.is_downlink:
        # the satellite's apoapsis is what we consider
        dist = model.range_to_geo_km
    else:
        dist = model.geo_altitude_km

    spreading = utils.spreading_loss_db(dist)
    return (model.peak_tx_eirp_dbw - spreading)


def _peak_pfd_at_geo_dbw_per_m2_per_hz(model):
    return utils.pfd_hz_manual_adjust(model.peak_pf_at_geo_dbw_per_m2,
                                      model.allocation_hz,
                                      1)


def _gso_victim_epfd(model):
    # http://famille.philippe.gerard.pagesperso-orange.fr/EPFD%20concept/Epfd%20conceptVfinale.htm
    # https://www.law.cornell.edu/cfr/text/47/25.208
    #
    # EPFD is, essentially, the PFD that would be required at the
    # victim's peak-gain angle to produce the same received power as
    # the sum of transmitters.
    #
    # In linear units:
    # epfd = sum(epfd_i) for all contributors
    #
    # We assume a single contributor, so:
    # Pr = pfd * Ar = epfd * Ar_max
    # epfd = pfd * Ar / Ar_max
    #
    # In log units:
    # epfd = pfd + Ar - Ar_max
    #
    # Alternatively, we can apply a correction term to the pfd
    # epfd_delta = Ar - Ar_max
    pass

def _rx_power_dbw(model):
    return (model.tx_eirp_dbw
            - model.total_channel_loss_db
            - model.polarization_mismatch_loss_db
            - model.rx_antenna_pointing_loss_db
            + model.rx_antenna_gain_dbi)


def _rx_antenna_effective_area_dbm2(model):
    # Sklar, Page 253, Equation 5.8
    return (model.rx_antenna_gain_dbi
            + utils.to_db(model.wavelength_m**2 / (4 * math.pi)))


def _rx_g_over_t_db(model):
    return model.rx_antenna_gain_dbi - model.rx_noise_temp_dbk


def _rx_n0_dbw_per_hz(model):
    return model.boltzmann_J_per_K_db + model.rx_noise_temp_dbk


def _cn0_db(model):
    return model.rx_power_dbw - model.rx_n0_dbw_per_hz


def _excess_noise_bandwidth_loss_db(model):
    req_bw = model.required_rx_bw_dbhz
    noise_bw = utils.to_db(model.rx_noise_bw_hz) if model.rx_noise_bw_hz else req_bw
    return noise_bw - req_bw


def _rx_eb(model):
    return model.rx_power_dbw - model.bitrate_dbhz


def _rx_ebn0_db(model):
    return model.rx_eb - model.rx_n0_dbw_per_hz


def _additional_rx_losses_db(model):
    return (model.implementation_loss_db
            + model.excess_noise_bandwidth_loss_db)


def _required_ebn0_db(model):
    return (model.required_demod_ebn0_db
            + model.additional_rx_losses_db)


def _link_margin_db(model):
    return model.rx_ebn0_db - model.required_ebn0_db


class LinkBudget(object):

    def __init__(self,
                 name='Link Budget',
                 is_downlink=True,
                 rx_antenna_noise_temp_k=300,
                 target_margin_db=1.0):

        self.tribute = {
            # calculators
            'pf_dbw_per_m2': _pf_dbw_per_m2,
            'peak_pfd_at_geo_dbw_per_m2_per_hz': _peak_pfd_at_geo_dbw_per_m2_per_hz,
            'rx_power_dbw': _rx_power_dbw,
            'rx_antenna_effective_area_dbm2': _rx_antenna_effective_area_dbm2,
            'rx_g_over_t_db': _rx_g_over_t_db,
            'rx_n0_dbw_per_hz': _rx_n0_dbw_per_hz,
            'cn0_db': _cn0_db,
            'excess_noise_bandwidth_loss_db': _excess_noise_bandwidth_loss_db,
            'pfd_dbw_per_m2_per_hz': _pfd_dbw_per_m2_per_hz,
            'rx_eb': _rx_eb,
            'rx_ebn0_db': _rx_ebn0_db,
            'link_margin_db': _link_margin_db,
            'peak_pf_at_geo_dbw_per_m2': _peak_pf_at_geo_dbw_per_m2,
            'boltzmann_J_per_K': 1.3806488e-23,
            'boltzmann_J_per_K_db': utils.to_db(1.3806488e-23),
            'peak_pf_at_geo_dbw_per_m2': _peak_pf_at_geo_dbw_per_m2,
            'peak_pf_dbw_per_m2': _peak_pf_dbw_per_m2,
            'canonical_pf_dbw_per_m2': _canonical_pf_dbw_per_m2,
            'range_to_geo_km': _range_to_geo_km,
            'rx_rf_chain': _rx_rf_chain,
            'tx_inline_losses_db': _tx_inline_losses_db,
            'tx_power_at_antenna_dbw': _tx_power_at_antenna_dbw,
            'tx_eirp_dbw': _tx_eirp_dbw,
            'peak_tx_eirp_dbw': _peak_tx_eirp_dbw,
            'required_ebn0_db': _required_ebn0_db,
            'additional_rx_losses_db': _additional_rx_losses_db,
            'peak_pfd_dbw_per_m2_per_hz': _peak_pfd_dbw_per_m2_per_hz,

            # constants
            'budget_name': name,
            'boltzmann_J_per_K': 1.3806488e-23,
            'boltzmann_J_per_K_db': utils.to_db(1.3806488e-23),
            'is_downlink': is_downlink,
            'rx_antenna_noise_temp_k': rx_antenna_noise_temp_k,
            'target_margin_db': target_margin_db,
            }



#!/usr/bin/python

import collections
import distutils.sysconfig
import jinja2
import matplotlib.pyplot as plt
import numpy as np
import os
import site

import pylink.utils as utils


class Figure(object):

    def __init__(self, title=None):
        self.title = title

    def to_latex(self, dname='.', fname=None):

        if not fname:
            fname = self.fname()

        return '''
  \\begin{figure}
    \\caption{%s}
    \\includegraphics[width=\\linewidth]{%s}
    \\label{fig::pfd::%s}
  \\end{figure}
        ''' % (self.caption(), os.path.join(dname, fname), self.label())

    def _plot_vs_el(self, y_func, dname='.', fname=None):
        m = self.model
        e = self.model.enum

        orig_el = m.override_value(e.min_elevation_deg)

        fig = plt.figure()
        if self.title:
            fig.suptitle(self.title)
        ax = fig.add_subplot(1, 1, 1)

        # Plot any limit lines
        if 'pfd_limits' in vars(self) and self.pfd_limits:
            x = [p[0] for p in self.pfd_limits]
            y = [p[1] for p in self.pfd_limits]
            ax.plot(x, y, color='r', linewidth=2, label='Limit')

        # Now plot the PFD curve
        x = np.linspace(0.0, 90.0, 90)
        y = np.linspace(0.0, 90.0, 90)

        for i in range(len(x)):
            m.override(e.min_elevation_deg, x[i])
            y[i] = y_func(i)

        ax.plot(x, y, color='b')
        ax.set_xlabel('Elevation Angle (degrees)')
        ax.set_ylabel(self.ylabel())

        # XXX: matplotlib is doing weird things if you set ylim this way
        # if 'pfd_limits' in vars(self) and self.pfd_limits:
        #     upper = max(max(y), self.pfd_limits[-1][1])
        #     lower = min(min(y), self.pfd_limits[0][1])

        #     delta = upper - lower
        #     upper += delta*0.1
        #     lower -= delta*0.1

        #     plt.ylim(lower, upper)
        # else:
        #     upper = max(y)
        #     lower = min(y)

        if not fname:
            fname = self.fname()
        path = os.path.join(dname, fname)
        fig.savefig(path, transparent=True)

        if orig_el is not None:
            m.override(e.min_elevation_deg, orig_el)
        else:
            m.reset(e.min_elevation_deg)


class CanonicalPFDFigure(Figure):
    """Power Flux Density figure for the canonical PFD vs Elevation

    The canonical figure assumes complete and even use of the
    spectrum, and it also tends to assume max eirp from the
    transmitter.  Becauses this model is really only used for
    compliance purposes, we ignore things like rain losses.

    This figure also assumes we are dealing with a space-to-earth
    link.  For NGSO to GSO, or earth-to-space, this figure is
    inapplicable.

     * Space-to-Earth
     * Boresight
     * Full BW Utilization
    """

    def __init__(self,
                 model,
                 bw=4e3,
                 pfd_limits=None,
                 title=None):
        """Creates a new canonical pfd figure.

        model -- The DAG model.
        bw -- BW for the PFD (defaults to 4kHz
        pfd_limits -- If it is PFD vs elevation, also plot these PFD limits
        """
        self.title = title
        self.model = model
        self.pfd_limits = pfd_limits
        self.bw = bw

    def label(self):
        return 'pfd-canonical-%d%s' % utils.human_hz(self.bw)

    def fname(self):
        return 'pfd-canonical-%d%s.png' % utils.human_hz(self.bw)

    def ylabel(self):
        return 'Boresight PFD (dBW/m^2/%d%s)' % utils.human_hz(self.bw)

    def caption(self):
        return "Peak PFD at Earth Station assuming full BW utilization"

    def plot(self, dname='.', fname=None):
        def __y(i):
            m = self.model
            return utils.pfd_hz_manual_adjust(m.canonical_pf_dbw_per_m2,
                                              m.allocation_hz,
                                              self.bw)
        return self._plot_vs_el(__y, dname=dname, fname=fname)


class ExpectedPFDFigure(Figure):
    """Power Flux Density figure for the expected case

    The expected case is much more complicated and includes things
    like antenna patterns, satellite pointing, and modulation code
    selection.  Typically, when a higher modulation code is selected,
    the BW utilization drops slightly resulting in slightly higher PFD
    because the total signal power is spread over slightly fewer Hz.
    As a result, this graph is not expected to be smooth.  It should,
    however, be monotonic.

    This figure also assumes we are dealing with a space-to-earth
    link.  For NGSO to GSO, or earth-to-space, this figure is
    inapplicable.

     * Space-to-Earth
     * Expected antenna gain
     * Expected BW utilization
    """

    def __init__(self,
                 model,
                 bw=4e3,
                 pfd_limits=None,
                 title=None):
        """Creates a new canonical pfd figure.

        model -- The DAG model.
        bw -- BW for the PFD (defaults to 4kHz
        pfd_limits -- If it is PFD vs elevation, also plot these PFD limits
        """
        self.title = title
        self.model = model
        self.pfd_limits = pfd_limits
        self.bw = bw

    def label(self):
        return 'pfd-expected-%d%s' % utils.human_hz(self.bw)

    def ylabel(self):
        return 'Expected PFD (dBW/m^2/%d%s)' % utils.human_hz(self.bw)

    def fname(self):
        return 'pfd-expected-%d%s.png' % utils.human_hz(self.bw)

    def caption(self):
        return "Expected PFD at Earth Station under normal operations"

    def plot(self, dname='.', fname=None):
        def __y(i):
            return utils.rx_pfd_hz_adjust(self.model,
                                          self.model.pf_dbw_per_m2,
                                          self.bw)
        return self._plot_vs_el(__y, dname=dname, fname=fname)


class PFDvsBWFigure(Figure):
    """Peak Power Flux Density figure vs BandWidth

    For compliance reasons, it can be useful to show a graph of the
    peak acheivable PFD vs BW for both the receiver and the GSO case.

     * Space-to-Earth or Earth-to-Space
     * Peak antenna gain
     * Full BW utilization
     * GSO or Receiver
    """

    def __init__(self,
                 model,
                 start_hz=1,
                 end_hz=4e3,
                 is_gso=False,
                 pfd_limits=None,
                 title=None):
        """Creates a new figure.

        model -- The DAG model.
        start_hz -- If so, starting BW
        end_hz -- And the ending BW
        is_gso -- True if looking at PFD at GSO, otherwise it is the receiver
        pfd_limits -- plot these PFD limits
        """

        self.title = title
        self.model = model
        self.start_hz = start_hz
        self.end_hz = end_hz
        self.is_gso = is_gso
        self.pfd_limits = pfd_limits

    def fname(self):
        prefix = 'gso' if self.is_gso else 'rx'
        prefix = 'pfd-' + prefix
        return '%s_%d-%d.png' % (prefix, self.start_hz, self.end_hz)

    def label(self):
        prefix = 'gso' if self.is_gso else 'rx'
        return '%s_%d-%d' % (prefix, self.start_hz, self.end_hz)

    def caption(self):
        if self.is_gso:
            return 'Peak PFD at GSO vs Bandwidth'
        else:
            if self.model.is_downlink:
                return 'Peak PFD at Surface vs Bandwidth'
            else:
                return 'Peak PFD at Receiver vs Bandwidth'

    def plot(self, dname='.', fname=None):
        m = self.model

        if self.is_gso:
            pf = m.peak_pf_at_geo_dbw_per_m2
        else:
            pf = m.peak_pf_dbw_per_m2

        fig = plt.figure()
        if self.title:
            fig.suptitle(self.title)
        ax = fig.add_subplot(1, 1, 1)
        n = int(self.end_hz) - int(self.start_hz)
        x = np.linspace(1, n, n-1)
        x += self.start_hz
        y = np.zeros(n-1)
        for i in range(len(x)):
            bw = x[i]
            y[i] = utils.pfd_hz_manual_adjust(pf, m.allocation_hz, bw)

        ax.plot(x, y, color='b', label='PFD (dBW/m^2)')

        # Plot any limit lines
        if self.pfd_limits:
            x = [p[0] for p in self.pfd_limits]
            y = [p[1] for p in self.pfd_limits]
            ax.plot(x, y, color='r', linewidth=2, label='Limit')


        ax.set_xlabel('Bandwidth (Hz)')
        if self.is_gso:
            ax.set_ylabel('PFD at GSO(dBW/m^2)')
        else:
            ax.set_ylabel('PFD at Receiver (dBW/m^2)')
        if not fname:
            fname = self.fname()
        path = os.path.join(dname, fname)
        fig.savefig(path, transparent=True)


class BitrateFigure(Figure):
    """Max bitrate vs Elevation Figure

    This graph assumes normal operations and indicates the max
    available bitrate.
    """

    def __init__(self, model, title):
        self.model = model
        self.title = title

    def label(self):
        return 'max-bitrate'

    def fname(self):
        return 'max-bitrate.png'

    def caption(self):
        return 'Bitrate with %ddB of Link Margin vs Elevation' % self.model.target_margin_db

    def ylabel(self):
        return 'Max Bitrate (MHz)'

    def plot(self, dname='.', fname=None):
        def __y(i):
            R = self.model.max_bitrate_hz
            m = self.model
            return self.model.max_bitrate_hz / 1.0e6
        return self._plot_vs_el(__y, dname=dname, fname=fname)


class Report(object):
    """LaTeX report Generator.

    Generates latex reports with great extensibility for either
    general engineering, or compliance reports for people like
    NOAA/NASA/FCC/ITU/ESA/TLA/ETLA.
    """

    def __init__(self, model):
        self.model = model
        self.enum = model.enum

    def _table(self, items, use_headers=True):
        m = self.model
        table = ''
        table += '\\begin{tabular}{p{3in}p{1in}l}\n'
        for section in items:
            s = section
            if section == 'Transmitter': s = 'Transmitter (%s)' % m.transmitter_name
            if section == 'Receiver': s = 'Receiver (%s)' % m.rx_name
            if use_headers:
                table += '\\textbf{%s} & & \\\\\n' % s
                table += '\\hline \\\\\n'
            lines = items[section]
            for line in lines:
                v = line[1]
                v = '%.2f' % v if isinstance(v, float) else '%s' % v
                table += '%s & %s & $%s$ \\\\\n' % (line[0],
                                                    v,
                                                    line[2])
            table += '\\\\\n'

        table += '\\end{tabular}\n'
        return table

    def _rx_chain_table(self):
        m = self.model
        e = m.enum
        chain = m.rx_rf_chain

        fields = [('Element', '', 1.25),
                  ('Gain', 'dB', .47),
                  ('Noise Figure', 'dB', .47),
                  #('Gain Factor', '', .5),
                  #('Noise Factor', '', .5),
                  ('Noise Temp.', 'K', .5),
                  ('Upstream Gain Factor', '', .6),
                  ('Contrib. Noise Temp', 'K', .5),
                  ('Contrib. Noise Factor', '', .5),
                  ('Noise Temp Pct Contrib.', '\\%', .75),
                  ]

        field_definitions = ''.join(['p{%sin}' % f[-1] for f in fields])
        table = ['\\hspace*{-0.75in} \\begin{tabular}{%s}' % field_definitions]

        # titles
        row = ['\\textbf{%s}' % f[0] for f in fields]
        table.append(' & '.join(row) + '\\\\')
        table.append('\hline \\\\')

        # Units
        row = []
        for f in fields:
            if len(f[1]):
                row.append('\\textbf{$%s$}' % (f[1]))
            else:
                row.append('')
        table.append(' & '.join(row) + '\\\\')
        table.append('\hline \\\\')


        for el in chain:
            pct = el.noise_temp_contrib / m.rx_system_noise_temp_k

            def __process_element(val):
                if isinstance(val, int):
                    return str(val)
                if val > 100:
                    return str(int(val))
                if val > 1000:
                    return '%g'%val
                return '%.2f' % val

            row = [el.gain_db,
                   el.noise_figure_db,
                   el.noise_temp_k,
                   el.prev_gain,
                   el.noise_temp_contrib,
                   el.noise_factor_contrib,
                   ] + [pct*100.0,]
            row = [el.name] + [__process_element(v) for v in row]
            row = (' & '.join(row)) + '\\\\'
            table.append(row)

        table.append('\\end{tabular}')
        return '\n'.join(table)

    def _humanize_hz(self, label, val):
        return (label,) + utils.human_hz(val)

    def _humanize_m(self, label, val):
        return (label,) + utils.human_m(val)

    def _static_pfd_entry(self, bw=0, is_gso=False):
        m = self.model

        if is_gso:
            retval = m.peak_pf_at_geo_dbw_per_m2
        else:
            retval = m.peak_pf_dbw_per_m2

        pfd_per_hz = retval - m.required_rx_bw_dbhz
        if bw:
            retval = pfd_per_hz + min(utils.to_db(bw), m.required_rx_bw_dbhz)
            (bw, bw_unit,) = utils.human_hz(bw)
            if is_gso:
                label = 'Peak PFD at GSO per %.2g%s' % (bw, bw_unit)
            elif m.is_downlink:
                label = 'Peak PFD at Surface per %.2g%s' % (bw, bw_unit)
            else:
                label = 'Peak PFD at Receiver per %.2g%s' % (bw, bw_unit)
            unit = 'dBW/m^2/%.2g%s' % (bw, bw_unit)
        else:
            if is_gso:
                label = 'Peak PFD at GSO'
            else:
                if m.is_downlink:
                    label = 'Peak PFD at Surface'
                else:
                    label = 'Peak PFD at Receiver'
            unit = 'dBW/m^2'
        return (label, retval, unit,)

    def _to_label(self, s):
        return self._file_namify(s)

    def _file_namify(self, s):
        for c in "\"'. \t":
            s = s.replace(c, '_')
        return s

    def to_latex(self,
                 fname,
                 author='',
                 intro='',
                 expo='',
                 rx_pfd_bws=[],
                 gso_pfd_bws=[],
                 pfd_figures=[],
                 added_sections=[],
                 added_interference_sections=[],
                 bitrate_figure=None,
                 watermark_text=None):
        """Export the budget to LaTeX

        fname -- Top level latex path
        author -- The author's name/entry in the header
        intro -- An optional section of latex to be included after the title
        expo -- An optional section of latex to be included at the end
        rx_pfd_bws -- List of PFD BW's to include in the interference section
        gso_pfd_bws -- List of PFD BW's to include in the interference section
        added_sections -- Additional sections to add before the budget section
        added_iterference_sections -- Additional subsections for interference
        pfd_figures -- Any desired PFD figure objects
        bitrate_figure -- Any desired bitrate figure objects
        watermark_text -- Watermark text (such as DRAFT or CONFIDENTIAL)

        Interference Subsections:
        [('Subsection Title', [
            ('Item Name', item_value, 'units'),
            ('Another Item Name', another_value, 'units'),
          ]),]

        Additional Sections:
        [('Section Name', [
            subsection,
            subsection,
          ]),]
        """
        
        m = self.model
        e = self.model.enum

        dname = os.path.dirname(fname)

        if intro is None:
            intro = ''

        top = self._jinja().get_template('budget.tex')

        orbit_items = collections.OrderedDict([
            ('Orbit', [
                ('Apoapsis', m.apoapsis_altitude_km, 'km'),
                ('Periapsis', m.periapsis_altitude_km, 'km'),
                ('Mean Altitude', m.mean_orbit_altitude_km, 'km'),
                ]),
            ('Constants', [
                ('Earth Radius', m.earth_radius_km, 'km'),
                ('GS Angle of Elevation', m.min_elevation_deg, '^{\\circ}'),
                ]),
            ('Key Outputs', [
                ('Slant Range', m.slant_range_km, 'km'),
                ('Satellite Angle from Boresight', m.satellite_antenna_angle_deg, '^{\\circ}'),
                ]),
            ])

        tx_items = collections.OrderedDict([
            ('Antenna', [
                ('Sampled Gain', m.tx_antenna_gain_dbi, 'dBi'),
                ('Antenna Pointing Loss', m.tx_antenna_pointing_loss_db, 'dB'),
                ('Gain at Boresight', m.tx_antenna_boresight_gain_dbi, 'dBi'),
                ('Average Nadir Gain', m.tx_antenna_average_nadir_gain_dbi, 'dBi'),
                ('Polarization', m.tx_antenna_polarization, ''),
                ]),

            ('Transmitter', [
                ('Transmit Power at PA', m.tx_power_at_pa_dbw, 'dBW'),
                ('Inline Losses', m.tx_inline_losses_db, 'dB'),
                ('Transmit Power at Antenna', m.tx_power_at_antenna_dbw, 'dBW'),
                ('EIRP', m.tx_eirp_dbw, 'dBW'),
                # EVM
                ]),
            ])

        interference_sections = [
            ('Transmitter', [
                ('Transmit Power at Antenna', m.tx_power_at_antenna_dbw, 'dBW'),
                ('Peak Antenna Gain', m.tx_antenna_peak_gain_dbi, 'dBi'),
                ('Peak EIRP', m.peak_tx_eirp_dbw, 'dBW'),
                ('Nadir-to-GS Gain', m.tx_antenna_gain_dbi, 'dBi'),
                ('Nadir-to-GS EIRP', m.tx_eirp_dbw, 'dBW'),
                ('Average Antenna Gain', m.tx_antenna_average_gain_dbi, 'dBi'),
                ]),
            ]

        pfd_lst = []
        for bw in rx_pfd_bws:
            pfd_lst.append(self._static_pfd_entry(bw=bw, is_gso=False))
        for bw in gso_pfd_bws:
            pfd_lst.append(self._static_pfd_entry(bw=bw, is_gso=True))
        if len(pfd_lst):
            interference_sections.append(('Peak Power Flux Density at Receiver', pfd_lst))
        interference_sections += added_interference_sections

        interference_items = collections.OrderedDict(interference_sections)

        channel_items = collections.OrderedDict([
            ('Constants', [
                self._humanize_hz('Center Frequency', m.center_freq_hz),
                ('Speed of Light', int(m.speed_of_light_m_per_s), 'm/s'),
                self._humanize_hz('Allocated Bandwidth', m.allocation_hz),
                self._humanize_hz('Start of Allocation', m.allocation_start_hz),
                self._humanize_hz('End of Allocation', m.allocation_end_hz),
                ('Wavelength',) + utils.human_m(m.wavelength_m),
                ]),

            ('Losses', [
                ('Unity Gain Propagation Loss', m.unity_gain_propagation_loss_db, 'dB'),
                ('Atmospheric Loss', m.atmospheric_loss_db, 'dB'),
                ('Ionospheric Loss', m.ionospheric_loss_db, 'dB'),
                ('Rain Loss', m.rain_loss_db, 'dB'),
                ('Multipath Fading Loss', m.multipath_fading_db, 'dB'),
                ('Total Channel Loss', m.total_channel_loss_db, 'dB'),
                ]),

            ('Modulation (%s)' % m.modulation_name, [
                ('Modulation Name', m.modulation_name, ''),
                ('Modulation Code', m.best_modulation_code.name, ''),
                ('Tx Spectral Efficiency', m.best_modulation_code.tx_eff, ''),
                self._humanize_hz('Bitrate', m.bitrate_hz),
                ('Bitrate', m.bitrate_dbhz, 'dBHz'),
                ('Required Demodulation $E_b/N_0$', m.required_demod_ebn0_db, 'dB'),
                self._humanize_hz('Required Demod Bandwidth', m.required_rx_bw_hz),
                ('Required Demod Bandwidth', m.required_rx_bw_dbhz, 'dBHz'),
                ('Required Transmit Bandwidth', m.required_tx_bw_dbhz, 'dBHz'),
                ]),
            ])

        rx_items = collections.OrderedDict([
            ('Antenna', [
                ('Noise Temperature', m.rx_antenna_noise_temp_k, 'K'),
                ('Sampled Gain', m.rx_antenna_gain_dbi, 'dBi'),
                ('Antenna Pointing Loss', m.rx_antenna_pointing_loss_db, 'dB'),
                ('Gain at Boresight', m.rx_antenna_boresight_gain_dbi, 'dBi'),
                ('Average Nadir Gain', m.rx_antenna_average_nadir_gain_dbi, 'dBi'),
                ('Polarization', m.rx_antenna_polarization, ''),
                ]),

            ('System', [
                ('Total Receiver Noise Temperature', m.rx_noise_temp_k, 'K'),
                ('Total Receiver Noise Temperature', m.rx_noise_temp_dbk, 'dBK'),
                ('Receive Chain Noise Temperature', m.rx_system_noise_temp_k, 'K'),
                ('Receive Chain Noise Factor', m.rx_system_noise_factor, ''),
                ('Receive Chain Noise Figure', m.rx_system_noise_figure, 'dB'),
                self._humanize_hz('Receiver Noise Bandwidth', m.bitrate_hz),
                ('Implementation Loss', m.implementation_loss_db, 'dB'),
                ]),
            ])

        budget_items = collections.OrderedDict([
            ('Transmitter', [
                ('Power at Antenna', m.tx_power_at_antenna_dbw, 'dBW'),
                ('Gain', m.tx_antenna_gain_dbi, 'dBi'),
                ('Antenna Pointing Loss', m.tx_antenna_pointing_loss_db, 'dB'),
                ('EIRP', m.tx_eirp_dbw, 'dBW'),
                ]),
                
            ('Channel', [
                self._humanize_hz('Center Frequency', m.center_freq_hz),
                ('Free Space Loss', m.unity_gain_propagation_loss_db, 'dB'),
                ('Total Channel Loss', m.total_channel_loss_db, 'dB'),
                ('Occupied Bandwidth', m.required_tx_bw_dbhz, 'dBHz'),
                ]),

            ('Receiver', [
                ('Signal Power Flux at RX Antenna', m.pf_dbw_per_m2, 'dBW/m^2'),
                ('Antenna Gain', m.rx_antenna_gain_dbi, 'dBi'),
                ('Antenna Effective Area', m.rx_antenna_effective_area_dbm2, 'dBm^2'),
                ('Polarization Mismatch Loss', m.polarization_mismatch_loss_db, 'dB'),
                ('RX System Noise Temperature', m.rx_noise_temp_k, 'K'),
                ('RX Figure of Merit ($G/T$)', m.rx_g_over_t_db, 'dB/K'),
                ('Noise Spectral Density ($N_0$)', m.rx_n0_dbw_per_hz, 'dBW/Hz'),
                ('Received Signal Power', m.rx_power_dbw, 'dBW'),
                ('Excess Noise BW Loss', m.excess_noise_bandwidth_loss_db, 'dB'),
                ('Implementation Loss', m.implementation_loss_db, 'dB'),
                ('Required Demodulation $E_b/N_0$', m.required_demod_ebn0_db, 'dB'),
                ('Antenna Pointing Loss', m.rx_antenna_pointing_loss_db, 'dB'),
                ('Carrier to Noise ($C/N_0$)', m.cn0_db, 'dBHz'),

                ]),

            ('Eb/N0', [
                ('BitRate', m.bitrate_dbhz, 'dBHz'),
                ('Energy per Bit ($E_b$)', m.rx_eb, 'dBW/Hz'),
                ('Received $E_b/N_0$', m.rx_ebn0_db, 'dB'),
                ('Required $E_b/N_0$', m.required_ebn0_db, 'dB'),
                ]),
            ('Key Outputs', [
                self._humanize_hz('Bitrate', m.bitrate_hz),
                ('Link Margin', m.link_margin_db, 'dB'),
                ]),
            ])

        rx_pattern_fname = '%sRXPattern.png' % self._file_namify(m.budget_name)
        m.rx_antenna_obj.plot_pattern(os.path.join(dname, rx_pattern_fname))

        tx_pattern_fname = '%sTXPattern.png' % self._file_namify(m.budget_name)
        m.tx_antenna_obj.plot_pattern(os.path.join(dname, tx_pattern_fname))

        if watermark_text:
            draft_mark = '''
\\newwatermark[allpages,color=black!10,angle=45,scale=3,xpos=0,ypos=0]{%s}
''' % watermark_text
# \\newwatermark[allpages,color=black!10,angle=45,scale=3,xpos=0,ypos=2in]{DRAFT}
        else:
            draft_mark = ''

        if len(expo) > 0:
            expo = '\\newpage\n' + expo

        figs = []
        for fig in pfd_figures:
            fig.plot(dname=os.path.dirname(fname))
            figs.append(fig.to_latex())
        interference_figures = '\n'.join(figs)

        if bitrate_figure:
            bitrate_figure.plot(dname=os.path.dirname(fname))
            budget_figures = bitrate_figure.to_latex()
        else:
            budget_figures = ''

        added_tables = []
        for section_name, section in added_sections:
            added_tables.append('''
\\newpage
\\section{%s}
\\label{section::%s}
\\begin{center}
''' % (section_name, self._to_label(section_name)))
            added_tables.append(self._table(collections.OrderedDict(section)))
            added_tables.append('\\end{center}')

        kwargs = {
            'intro':intro,
            'expo':expo,
            'budgetTitle': m.budget_name,
            'table': self._table(budget_items),
            'interferenceFigures': interference_figures,
            'txPatternFname': tx_pattern_fname,
            'rxPatternFname': rx_pattern_fname,
            'draftMark': draft_mark,
            'orbitTable': self._table(orbit_items),
            'transmitterTable': self._table(tx_items),
            'channelTable': self._table(channel_items),
            'receiverTable': self._table(rx_items),
            'rxChainTable': self._rx_chain_table(),
            'interferenceTable': self._table(interference_items),
            'author': author,
            'addedSections': '\n'.join(added_tables),
            'budgetFigures': budget_figures,
            }
        with open(fname, 'w') as fd:
            fd.write(top.render(**kwargs))


    def _jinja(self, basedir=None):
        if not basedir:
            lib = distutils.sysconfig.get_python_lib()
            rel = 'pylink/tex'
            basedir = os.path.join(lib, rel)

        env = jinja2.Environment(
            block_start_string = '\BLOCK{',
            block_end_string = '}',
            variable_start_string = '\VAR{',
            variable_end_string = '}',
            comment_start_string = '\#{',
            comment_end_string = '}',
            line_statement_prefix = '%%',
            line_comment_prefix = '%#',
            trim_blocks = True,
            autoescape = False,
            loader = jinja2.FileSystemLoader(basedir))
        return env



#!/usr/bin/python


class TaggedAttribute(object):
    """Simple container object to tag values with attributes.

    Feel free to initialize any node with a TA instead of its actual
    value only and it will then have the desired metadata.  For example:

    from pylink import TaggedAttribute as TA
    tx_power = TA(2,
                  part_number='234x',
                  test_report='http://reports.co/234x')
    m = DAGModel([pylink.Transmitter(tx_power_at_pa_dbw=tx_power)])
    """

    def __init__(self, value, **kwargs):
        self.meta = kwargs
        self.value = value



#!/usr/bin/python

import argparse
import numpy as np
import pylink
import math
import jinja2
import matplotlib.pyplot as plt


def _interpolate(tgt, vals, off):
    def __search(tgt, vals):
        low = 0
        hi = len(vals)
        cur = int(hi/2)
        while (hi - low) > 1:
            if vals[cur][0] > tgt:
                hi = cur
                cur = int(low + (cur - low)/2)
            elif vals[cur][0] < tgt:
                low = cur
                cur = int(cur + (hi - cur)/2)
            elif vals[cur][0] == tgt:
                return cur
        return low

    lo = __search(tgt, vals)

    # correct value
    if vals[lo][0] == tgt:
        return vals[lo][off]

    hi = lo + 1
    x0 = vals[lo][0]
    x1 = vals[hi][0]
    y0 = vals[lo][off]
    y1 = vals[hi][off]
    rate = (y1-y0)/(x1-x0)
    return rate*(tgt-x0)+y0


class HyperSpectralSNRBudget(object):

    def __init__(self,
                 orbital_solar_irradiance,
                 ground_solar_irradiance,
                 reflectance_db=pylink.to_db(0.5),
                 fore_optics_efficiency_db=pylink.to_db(0.95),
                 focusing_optics_efficiency_db=pylink.to_db(0.95),
                 grating_efficiency_db=pylink.to_db(0.8),
                 sensor_quantum_efficiency_db=pylink.to_db(0.8),
                 read_out_noise_e=50,
                 pixel_pitch_um=15,
                 fwhm_nm=10,
                 gsd_m=15,
                 gmc=1,
                 lambda_nm=1990,
                 lens_radius_m=0.3,
                 bits_per_sample=12,
                 spatial_channels=640):

        self.tribute = {
            # Constants
            'orbital_solar_irradiance_w': orbital_solar_irradiance,
            'ground_solar_irradiance_w': ground_solar_irradiance,
            'speed_of_light_m_s': 299792458,
            'plancks_constant': 6.626068*10**-34,
            'grav_const': 6.67*10**-11,
            'earth_mass_kg': 5.98*10**24,
            'reflectance_db': reflectance_db,
            'fore_optics_efficiency_db': fore_optics_efficiency_db,
            'grating_efficiency_db': grating_efficiency_db,
            'focusing_optics_efficiency_db': focusing_optics_efficiency_db,
            'sensor_quantum_efficiency_db': sensor_quantum_efficiency_db,
            'read_out_noise_e': read_out_noise_e,
            'pixel_pitch_um': pixel_pitch_um,
            'fwhm_nm': fwhm_nm,
            'gsd_m': gsd_m,
            'gmc': gmc,
            'lambda_nm': lambda_nm,
            'lens_radius_m': lens_radius_m,
            'bits_per_sample': bits_per_sample,
            'spatial_channels': spatial_channels,

            # Calculators
            'incident_power_flux_density_dbw_m2_nm': self._incident_power_flux_density_dbw_m2_nm,
            'reflected_power_flux_density_dbw_m2_nm': self._reflected_power_flux_density_dbw_m2_nm,
            'reflected_power_density_dbw_nm': self._reflected_power_density_dbw_nm,
            'reflected_power_dbw': self._reflected_power_dbw,
            'power_flux_at_lens_dbw_m2': self._power_flux_at_lens_dbw_m2,
            'received_power_dbw': self._received_power_dbw,
            'orbital_velocity_km_per_s': self._orbital_velocity_km_per_s,
            'orbital_period_s': self._orbital_period_s,
            'shutter_time_s': self._shutter_time_s,
            'energy_at_sensor_dbj': self._energy_at_sensor_dbj,
            'signal_electrons_in_well_dbe': self._signal_electrons_in_well_dbe,
            'noise_electrons_dbe': self._noise_electrons_dbe,
            'snr_db': self._snr_db,
            'ground_area_dbm2': self._ground_area_dbm2,
            'optical_loss_db': self._optical_loss_db,
            'atmospheric_loss_db': self._atmospheric_loss_db,
            }

    def _incident_power_flux_density_dbw_m2_nm(self, model):
        PFD = max(_interpolate(model.lambda_nm,
                               model.ground_solar_irradiance_w,
                               1),
                  1e-10)
        return pylink.to_db(PFD)

    def _reflected_power_flux_density_dbw_m2_nm(self, model):
        # Incoming Power Flux
        Pi = model.incident_power_flux_density_dbw_m2_nm

        # Lambertian Reflectance (dimensionless)
        Rl = model.reflectance_db

        # Reflected Power Flux
        Pr = Pi + Rl

        return Pr

    def _ground_area_dbm2(self, model):
        return pylink.to_db(model.spatial_channels * model.gsd_m**2)

    def _reflected_power_density_dbw_nm(self, model):
        return (0.0
                + model.reflected_power_flux_density_dbw_m2_nm
                + model.ground_area_dbm2)

    def _reflected_power_dbw(self, model):
        return (1.0
                + model.reflected_power_density_dbw_nm
                + pylink.to_db(model.fwhm_nm))

    def _atmospheric_loss_db(self, model):
        # Power incident on the atmosphere
        Pi = _interpolate(model.lambda_nm,
                          model.orbital_solar_irradiance_w,
                          1)
        Pi = pylink.to_db(Pi)

        # Power transmitted through the atmosphere
        Pt = model.incident_power_flux_density_dbw_m2_nm

        return Pi - Pt

    def _power_flux_at_lens_dbw_m2(self, model):
        # Approximate the pixel as a point source and apply spreading loss
        # The approsimate transmission gain is 3dB as lambertian reflectors
        D = model.slant_range_km*1000
        spreading_loss_db = pylink.to_db(math.pi*2*D**2)
        return (0.0
                + model.reflected_power_dbw
                - spreading_loss_db
                - model.atmospheric_loss_db)

    def _received_power_dbw(self, model):
        # Incident Power Flux
        Pi = model.power_flux_at_lens_dbw_m2

        # Collecting Aperture Area
        A = pylink.to_db(math.pi * model.lens_radius_m**2)

        return Pi + A

    def _orbital_velocity_km_per_s(self, model):
        G = model.grav_const
        M = model.earth_mass_kg
        Re = model.earth_radius_km
        Alt = model.mean_orbit_altitude_km
        R = 1000 * (Re + Alt)
        return (G * M / R)**.5 / 1000.0

    def _orbital_period_s(self, model):
        v = model.orbital_velocity_km_per_s * 1000
        Re = model.earth_radius_km
        Alt = model.mean_orbit_altitude_km
        R = 1000 * (Re + Alt)
        return 2 * math.pi * R / v

    def _shutter_time_s(self, model):
        # Great Circle Circumference (m)
        Gc = model.earth_radius_km * 1000 * math.pi * 2

        # Orbital Period
        To = model.orbital_period_s

        # Ground Speed
        Gs = Gc / To

        # Time required to traverse the GSD
        T = model.gsd_m / Gs

        # GMC Factor
        return T * model.gmc

    def _optical_loss_db(self, model):
        # These efficiencies are all negative numbers, and this
        # function is meant to return a positive number (ie dB's of
        # loss as opposed to dB's of transmission), so we multiply by
        # negative 1.
        return -1 * (0.0
                     + model.fore_optics_efficiency_db
                     + model.grating_efficiency_db
                     + model.focusing_optics_efficiency_db)

    def _energy_at_sensor_dbj(self, model):
        # Energy striking the primary lens
        P_lens = model.received_power_dbw + pylink.to_db(model.shutter_time_s)

        # Optical losses associated with inefficiencies
        # Losses associated with the focal plane sensor
        P_sensor = P_lens - model.optical_loss_db

        return P_sensor

    def _signal_electrons_in_well_dbe(self, model):
        E_sensor = model.energy_at_sensor_dbj
        E_pixel = E_sensor - pylink.to_db(model.spatial_channels)
        P_r = E_pixel + model.sensor_quantum_efficiency_db
        lambda_dbm = pylink.to_db(model.lambda_nm) - 90
        h_db = pylink.to_db(model.plancks_constant)
        c_db = pylink.to_db(model.speed_of_light_m_s)
        return P_r + lambda_dbm - h_db - c_db

    def _noise_electrons_dbe(self, model):
        N = (0.0
             + model.read_out_noise_e
             + pylink.from_db(model.signal_electrons_in_well_dbe/2))
        return pylink.to_db(N)

    def _snr_db(self, model):
        return model.signal_electrons_in_well_dbe - model.noise_electrons_dbe



#!/usr/bin/python

from enum import Enum
import math
import numpy as np


def sequential_enum(*sequential, **named):
    """Returns a new enum with all given named and sequential nodes.

    sequential -- [str, str, ...]
    named -- str:int
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)


def node_associations(enum):
    """Returns dicts associating nodes -> names and vice versa from an enum.

    Given an enum, returns the dict of names to numbers and numbers to
    names.
    """
    node_to_name = {}
    name_to_node = {}
    for key, value, in enum.__dict__.items():
        if not key.startswith('_'):
            node_to_name[value] = key
            name_to_node[key] = value
    return (node_to_name, name_to_node)


def to_db(v):
    """linear to dB
    """
    return math.log(float(v), 10) * 10


def from_db(v):
    """dB to linear
    """
    return 10**(float(v)/10.0)


def human_hz(v):
    """Returns a number of Hz autoselected for Hz, kHz, MHz, and GHz
    """
    if v < 1e3:
        return (v, 'Hz')
    if v < 1e6:
        return (v/1.0e3, 'kHz')
    if v < 1e9:
        return (v/1.0e6, 'MHz')
    return (v/1.0e9, 'GHz')


def human_m(v):
    """Returns a distance autoselected for mm, cm, m, and km
    """
    if v < 1e-2:
        return (v*1.0e3, 'mm')
    if v < 1:
        return (v*1.0e2, 'cm')
    if v < 1000:
        return (v, 'm')
    return (v/1.0e3, 'km')


def human_b(bytes):
    """Returns a number of bytes autoselected for kB, MB, GB, TB

    Just to be clear, this will use the 1024 versions, not the 1000
    versions.
    """
    suffixes = ['TB', 'GB', 'MB', 'kB']
    for i in range(4):
        m = 1<<(10*(4-i))
        if bytes > m:
            return (bytes / float(m), suffixes[i])
    return (bytes, 'B')


def spreading_loss_db(dist_km):
    """Returns the loss, in dB, associated with distance alone.
    """
    return to_db(4 * math.pi * (dist_km * 1000)**2)


def rx_pfd_hz_adjust(model, base, n):
    """Transforms a PF into a PFD at N Hz for Rx
    """
    return pfd_hz_manual_adjust(base, model.required_rx_bw_hz, n)


def tx_pfd_hz_adjust(model, base, n):
    """Transforms a PF into a PFD at N Hz for Tx
    """
    return pfd_hz_manual_adjust(base, model.required_tx_bw_hz, n)


def pfd_hz_manual_adjust(base, occ, n):
    """Transforms a PF into a PFD at N Hz for Rx assuming occ BW
    """

    n_db = to_db(n)
    occ_db = to_db(occ)
    if n_db > occ_db:
        return base
    else:
        return base - occ_db + n_db


def pattern_generator(peak_gain_dbi, null=-20.0, eff=0.7):
    """Generates a sample antenna pattern.

    The pattern will be a main lobe, and the rest will be the <null>
    value.

    FIXME: See if a reasonable pattern, including side-lobes, can be
           generated easily

    peak_gain_dbi -- float
    null -- float, value outside of the main lobe
    eff -- float, antenna efficiency value
    """
    gain = peak_gain_dbi # just want it called peak_gain_dbi in docs

    # http://www.phys.hawaii.edu/~anita/new/papers/militaryHandbook/antennas.pdf
    X = 75146.0 * (eff-.7) + 41253.0
    bw_3db = (X / gain)**.5
    bw_null = bw_3db*1.5

    # determine step size
    min_n_steps = 25
    step = min(bw_null / min_n_steps, 1.0)
    n = 4.0 * int((360.0 / step) / 4.0) # ensure it is an even multiple of 4
    n_q = int(n / 4)

    def __sigmoid_gen(a=1.0, b=1.0, t=0.0):
        def retval(x):
            return to_db(a / (1.0 + math.exp(-1.0 * b * (x-t)))) + gain
        return retval

    # find the intercept with the null
    sig_upper = 4.0
    sig_lower = -6.0
    a = 1.0
    b = 1.0
    t = (math.log((a/(from_db(null-gain))) - 1) / b) + sig_lower

    sigmoid = np.vectorize(__sigmoid_gen(a=a, b=b, t=t), otypes=[np.float])

    # set things up
    n_half = int(bw_null / step / 2.0)
    sig_step = (sig_upper - sig_lower) / n_half
    main_g_r = np.arange(sig_lower, sig_upper, sig_step)# main lobe, gain, right
    main_a_r = np.arange(0, bw_null, n_half) # main lobe, angles, right
    main_g_r = sigmoid(main_g_r)
    delta = np.arange(0.0, n_half)
    delta *= (gain - main_g_r[-1]) / n_half
    main_g_r += delta

    # combine the left and right sides of the lobe
    main_g_l = list(main_g_r[:])
    main_g_l.reverse()
    main = list(main_g_r) + [gain] + main_g_l

    # combine the main lobe with null for the rest of the pattern
    retval = main + [null]*int(n-len(main))

    # find the peak
    n_rot = 0
    peak = min(retval)
    for i in range(len(retval)):
        if retval[i] > peak:
            n_rot = i
            peak = retval[i]

    # rotate the pattern
    retval = retval[n_rot:] + retval[:n_rot]

    return retval


def eirp_dbw_to_e_field_v_per_m(eirp_dbw, dist_m):
    """Returns the E field strength at a given distance/eirp

    The resulting value is in volts per meter, linear units.  The EIRP
    should be in dBW.
    """
    # http://www.ti.com/lit/an/swra048/swra048.pdf
    # https://www.craf.eu/useful-equations/conversion-formulae/
    # (PG)/(4*pi*d^2) = (E^2)/(120*pi)
    # eirp/(4*pi*d^2) = (E^2)/(120*pi)
    # E^2 = eirp * 120 * pi / 4 * pi * d^2
    # E^2 = eirp * 30 / d^2
    # E = (eirp * 30 / d^2) ^0.5
    # E = (eirp * 30) ^ 0.5 / d
    # E = (eirp * 30) ^ 0.5 / d

    E_db = ((eirp_dbw + to_db(30)) / 2) - to_db(dist_m)
    E = from_db(E_db)
    return E


def e_field_to_eirp_dbw(E, dist_m):
    """Returns the EIRP required to create a given E field strength.

    E field strength is in linear units, volts per meter.  The
    resulting value is in dBW.
    """
    # http://www.ti.com/lit/an/swra048/swra048.pdf
    # eirp = to_db(E^2 * dist_m^2 / 0.03)
    # https://www.craf.eu/useful-equations/conversion-formulae/
    # (PG)/(4*pi*d^2) = (E^2)/(120*pi)
    # eirp/(4*pi*d^2) = (E^2)/(120*pi)
    # eirp = (E^2) * (4*pi*d^2) / (120*pi)
    # eirp = (E^2 * d^2 * 4 * pi) / (120 * pi)
    # eirp = (E^2 * d^2) / 30

    eirp = E**2 * dist_m**2 / 30
    return to_db(eirp)



#!/usr/bin/python

import inspect
import math
import os
import pprint
import re
import sys
import tempfile
import traceback

import pylink.utils as utils

from pylink.tagged_attribute import TaggedAttribute


class LoopException(Exception):
    pass


class DAGModel(object):
    """DAG Solver

    After instantiating this class with a series of nodes, those nodes
    (and their recursively-calculated values) can be referenced as
    instance variables.  It also includes a solver, and the ability to
    override values.
    """

    def __init__(self, contrib=[], **extras):
        """Creates a new DAG Model

        contrib -- Iterable of tributaries
        extras -- Any additional node to add/override

        Please note that nodes are processed in the order in which
        they are received.  That means that if you add your node in
        the last tributary, it'll override any previous definition of
        that node.  Similarly, if you add your node in the kwargs,
        you'll override any node definition in the tributaries.
        """

        # Make sure the input looks sane
        for t in contrib:
            try:
                t.tribute
            except AttributeError:
                msg = ("One of your Tributaries doesn't have "
                       + "'tribute' defined: %s" % t.__class__.__name__)
                raise AttributeError(msg)

        # calculate the list of node names & enum
        names = []
        for mod in contrib:
            names += mod.tribute.keys()
        names.extend(extras.keys())
        self.enum = utils.sequential_enum(*names)

        # associate names to nodes
        (self._names, self._nodes,) = utils.node_associations(self.enum)

        # merge the contributions
        self._calc = {}
        self._values = {}
        self._meta = {}
        tributes = [m.tribute for m in contrib]
        tributes.append(extras)
        for t in tributes:
            self.accept_tribute(t)

        # Record the calculation stack for dependency tracking
        self._stack = []

        self._init_cache()

        # We start with an empty dependency tree and update as able
        self._deps = {}
        self._map_dependencies()

    def accept_tribute(self, t):
        for name, v, in t.items():
            node = self._nodes[name]
            if hasattr(v, '__call__'):
                self._calc[node] = v
            elif isinstance(v, TaggedAttribute):
                self._meta[node] = v.meta
                self._values[node] = v.value
            else:
                self._values[node] = v

    def clear_cache(self):
        """Clears the cache.

        Useful if you do something like accept_tribute().  Currently
        only used in testing.
        """
        self._init_cache()

    def is_calculated_node(self, node):
        """Determines whether or not the given node is calculated.
        """
        return node in self._calc

    def is_static_node(self, node):
        """Determines whether or not the given node is static.
        """
        return not self.is_calculated_node(node)

    def is_overridden(self, node):
        """Determines whether or not the given node is overridden.
        """
        return self.is_calculated_node(node) and node in self._values

    def get_meta(self, node):
        """Returns the metadata dict associated with this node
        """
        if node in self._meta:
            return self._meta[node]
        return None

    def set_meta(self, node, k, v):
        """Sets the desired key/value in this node's metadata dict
        """
        self._meta.setdefault(node, {k:v})
        self._meta[node][k] = v

    def _map_dependencies(self):
        # nodes => list of direct dependencies
        # self._deps is maintained by the cache system
        self._dep_names = self._named_deplist(self._deps)

        # nodes => flattened dependency list
        self._flat_deps = self._flattened_deps()
        self._flat_dep_names = self._named_deplist(self._flat_deps)

        # nodes => flattened list of nodes depending upon it
        self._clients = self._client_list(self._flat_deps)
        self._client_names = self._named_deplist(self._clients)

        self._deps_are_stale = False

    def _init_cache(self):
        self._cache = {}

    def _record_parent(self, node):
        if len(self._stack):
            dep = self._stack[-1]
            self._add_dependency_impl(node, dep)

    def print_dependencies(self):
        """Pretty Prints the dependency information for all nodes.

        First is the one-layer dependencies (ie link_margin_db depends
        upon required_ebn0_db and received_ebn0_db).

        Next the flattened dependencies are printed (so link_margin_db
        ends up depending upon lots and lots of things).

        Finally the flattened reverse dependencies are printed (we
        need these for proper cache invalidation).
        """
        if self._deps_are_stale:
            # This call is quite expensive, so we only want to do so
            # if necessary.
            self._map_dependencies()
        pprint.pprint(self._dep_names)
        pprint.pprint(self._flat_dep_names)
        pprint.pprint(self._client_names)

    def _add_dependency_impl(self, node, dep):
        self._deps.setdefault(dep, {node:0})
        self._deps[dep].setdefault(node, 0)
        self._deps[dep][node] += 1
        if self._deps[dep][node] == 1:
            self._deps_are_stale = True

    def cached_calculate(self, node, clear_stack=False):
        """Either return the cached value, or calculate/lookup the node's value.

        You really only need this method if you're introducing a
        cycle.

        clear_stack -- This kwarg permits the calculation with an
        empty stack.  That way the cycle-checker doesn't complain, and
        you can happily introduce cycles.  See the package README for
        more information about how to do this safely.
        """
        if clear_stack:
            orig_stack = self._stack
            self._stack = []

        self._record_parent(node)
        if node in self._cache:
            retval = self._cache[node]
        else:
            retval = self._calculate(node)

        if clear_stack:
            self._stack = orig_stack
        return retval


    def _calculate(self, node, stack=None):
        if stack:
            orig_stack = self._stack
            self._stack = stack

        if node in self._stack:
            stack = self._stack + [node]
            stack = [self.node_name(n) for n in stack]
            s = pprint.pformat(stack)
            raise LoopException("\n=== LOOP DETECTED ===\n%s" % s)

        self._stack.append(node)

        if node in self._values:
            retval = self._values[node]
        else:
            retval = self._calc[node](self)

        self._cache_put(node, retval)
        self._stack.pop()

        if stack:
            self._stack = orig_stack

        return retval

    def _cache_clear(self, node=None):
        if self._deps_are_stale:
            # This call is quite expensive, so we only want to do so
            # if necessary.
            self._map_dependencies()

        if node is not None:
            if node in self._cache:
                del self._cache[node]
            for client in self._clients[node]:
                if client in self._cache:
                    del self._cache[client]
        else:
            self._init_cache()

    def _cache_put(self, node, value):
        self._cache[node] = value

    def __getattr__(self, name):
        if name in self._nodes:
            node = self.node_num(name)
            if node not in self._values and node not in self._calc:
                name = self.node_name(node)
                msg = "It looks like you're missing an item: %s" % name
                raise AttributeError(msg)
            return self.cached_calculate(node)
        raise AttributeError("It looks like you're missing a node: %s" % name)

    def _client_list(self, flat_deps):
        retval = {}
        for dep in self._names:
            retval[dep] = []
            for node in flat_deps:
                if dep in flat_deps[node]:
                    retval[dep].append(node)
        return retval

    def _named_deplist(self, deps):
        retval = {}
        for node in deps:
            n = self.node_name(node)
            retval[n] = [ self.node_name(x) for x in deps[node] ]
        return retval

    def __flattened_deps_r(self, node, res={}):
        if node in self._deps:
            # We only register dependencies for dependent items
            for dep in self._deps[node]:
                res[dep] = 1
                self.__flattened_deps_r(dep, res=res)
        return res.keys()

    def _flattened_deps(self):
        retval = {}
        for node in self._deps:
            retval[node] = self.__flattened_deps_r(node, res={})
        return retval

    def node_name(self, node):
        """Returns the name of the node

        node -- this is the integer from the enum
        """    
        return self._names[node]

    def node_num(self, name):
        """Returns the node number from the name
        """    
        return self._nodes[name]

    def override(self, node, value):
        """Overrides a given node's value.

        If it's a static node, it redefines it.  If it's a calculated
        node it'll serve this static value instead of executing node.
        """
        self._cache_clear(node=node)
        self._values[node] = value

    def revert(self, node):
        """Reverts an override on a node.

        Please note that this operation only makes sense if you're
        reverting an override on a calculator.
        """
        if node in self._values:
            if node in self._calc:
                self._cache_clear(node=node)
                del self._values[node]
            else:
                name = self.node_name(node)
                msg = "You can't revert a static value: %s" % name
                raise AttributeError(msg)

    def override_value(self, node):
        """Returns the override value for a node.

        If you override a calculated node, this method will return the
        value to which it was overridden, otherwise None.  And if
        you're overriding the calculated node to return None, well,
        you're out of luck.
        """
        if node in self._values:
            return self._values[node]
        return None

    def _solve_for(self, var, fixed, fixed_value, start, stop, step):

        # The output variable should always be reverted
        self.revert(fixed)

        best_val = start
        best_diff = abs(fixed_value - self.cached_calculate(fixed))

        for i in range(0, int((stop-start)/step), 1):
            val = start + step*i
            self.override(var, val)
            diff = abs(fixed_value - self.cached_calculate(fixed))
            if diff < best_diff:
                best_diff = diff
                best_val = val

        return best_val

    def solve_for(self, var, fixed, fixed_value, start, stop, step, rounds=3):
        """Solve for a fixed variable by varying another.

        Using multiple <rounds>, this method solves for <var> by
        searching for the value that results in <fixed> being closest
        to <fixed_value> between <start> and <stop> at intervals of
        <step>.  Subsequent rounds will use the same number of steps
        though the step size will, of course, shrink.  After it will
        search within the winning range at higher precision.

        This method only works for either monotonic functions.  If
        there are two values that satisfy the constraint, it will find
        the one closest to <start>.

        var -- Node number to solve for (from the enum)
        fixed -- Node number constraining the search
        fixed_value -- The target value for the fixed node
        start -- The value for <var> at which to start
        stop -- The value for <var> at which to stop
        step -- The delta between points to try in the search
        rounds -- The total number of rounds to attempt
        """

        # It only makes sense to use a fixed variable that is a
        # calculator, so we check for that here.
        if not self.is_calculated_node(fixed):
            raise AttributeError("Can only solve with calculated outputs")

        # The number of rounds should be a member of Z+ > 0
        if int != type(rounds) or rounds < 1:
            raise AttributeError("Gimme an int number of rounds > 0, please.")

        # The control and response variables cannot be the same
        if fixed == var:
            raise AttributeError("Fixed and Variable nodes cannot be the same")

        # Preserve the original values
        def _static_val(node, is_calc):
            if is_calc: return self.override_value(node)
            return self.cached_calculate(node)
        fix_is_calc = self.is_calculated_node(fixed)
        fix_is_over = self.is_overridden(fixed)
        fix_orig = _static_val(fixed, fix_is_calc)
        var_is_calc = self.is_calculated_node(var)
        var_is_over = self.is_overridden(var)
        var_orig = _static_val(var, var_is_calc)

        n = (stop - start) / step
        for i in range(rounds):
            retval = self._solve_for(var,
                                     fixed, fixed_value,
                                     start,
                                     stop,
                                     step)
            ostep = step
            step = step/n
            if 0 == step:
                break
            start = max(retval - ostep/2.0 - 2*step, start)
            stop = min(retval + ostep/2.0 + 2*step, stop)


        # Restoration action truth table
        #
        # override, calculated, action
        #    0           0      restore
        #    0           1      nothing
        #    1           0      ERROR (can only override calculated nodes)
        #    1           1      restore

        restore_fix = not (fix_is_over ^ fix_is_calc)
        restore_var = not (var_is_over ^ var_is_calc)

        if restore_fix:
            self.override(fixed, fix_orig)

        if restore_var:
            self.override(var, var_orig)

        return retval

    def nodes(self):
        return self._names.keys()

    def export_deps_dot(self):
        retval = ['digraph {']

        for node, deps in self._deps.items():
            for dep in deps:
                retval.append('  %s -> %s' % (self.node_name(node),
                                              self.node_name(dep)))
        retval.append('}')
        return '\n'.join(retval)

    def export_deps_png(self, path):
        dot = self.export_deps_dot()
        n, dotpath = tempfile.mkstemp()
        fd = open(dotpath, 'w')
        fd.write(dot)
        fd.close()

        # build the graph
        # subprocess.call(['gv', '-o', path, dotpath])

        os.unlink(dotpath)



#!/usr/bin/python

import math

from .. import utils
from .. import element


def _rx_noise_temp_k(model):
    return model.rx_system_noise_temp_k + model.rx_antenna_noise_temp_k


def _rx_system_noise_temp_k(model):
    prev_gain = 1
    retval = 0

    # http://www.microwaves101.com/encyclopedias/noise-figure
    for el in model.rx_rf_chain:

        # Inputs
        el.gain = utils.from_db(el.gain_db)
        el.noise_factor = utils.from_db(el.noise_figure_db)

        el.noise_temp_k = (el.noise_factor-1)*model.room_temp_k

        el.accum_gain = prev_gain * el.gain
        el.noise_temp_contrib = el.noise_temp_k / prev_gain

        # Friis Cascade Equation: F_12 = F1 + (F2-1)/G_1
        # F_1:n = F_n-1 + (F_n-1)/G_1:n-1
        # F_
        el.noise_factor_contrib = (el.noise_factor-1) / prev_gain
        el.prev_gain = prev_gain

        retval += el.noise_temp_contrib
        prev_gain = el.accum_gain

    return retval


def _rx_noise_temp_dbk(model):
    return utils.to_db(model.rx_noise_temp_k)


def _rx_system_noise_factor(model):
    # observing this value ensures that the dependent values are
    # computed and stored
    unused = model.rx_system_noise_temp_k

    return sum([el.noise_factor_contrib for el in model.rx_rf_chain]) + 1


def _rx_system_noise_figure(model):
    return utils.to_db(model.rx_system_noise_factor)


class Receiver(object):

    def __init__(self,
                 rf_chain=None,
                 noise_bw_khz=None,
                 ground_noise_temp_k=290,
                 sky_noise_temp_k=70,
                 room_temp_k=290,
                 implementation_loss_db=0,
                 name='Receiver'):

        if rf_chain is None:
            # RADFACE
            balun = element.Element(name='Balun',
                                    gain_db=-0.9,
                                    noise_figure_db=0.9)
            coax = element.Element(name='Coax',
                                   gain_db=-0.24,
                                   noise_figure_db=0.24)
            lpf = element.Element(name='LFCN',
                                  gain_db=-0.8,
                                  noise_figure_db=0.8)
            sw = element.Element(name='Switch',
                                 gain_db=-0.4,
                                 noise_figure_db=0.4)
            lna = element.Element(name='LNA',
                                  gain_db=22,
                                  noise_figure_db=0.65)
            saw = element.Element(name='SAW Filter',
                                  gain_db=-1.7,
                                  noise_figure_db=1.7)
            rfic = element.Element(name='RFIC',
                                   gain_db=10,
                                   noise_figure_db=6)
            rf_chain = [balun, coax, lpf, sw, lna, saw, rfic]

        noise_bw_hz = noise_bw_khz * 1000 if noise_bw_khz else None

        self.tribute = {
            # calculators
            'rx_noise_temp_k': _rx_noise_temp_k,
            'rx_system_noise_temp_k': _rx_system_noise_temp_k,
            'rx_system_noise_factor': _rx_system_noise_factor,
            'rx_system_noise_figure': _rx_system_noise_figure,
            'rx_noise_temp_dbk': _rx_noise_temp_dbk,

            # constants
            'rx_name': name,
            'receiver_rf_chain': rf_chain,
            'rx_noise_bw_hz': noise_bw_hz,
            'sky_noise_temp_k': sky_noise_temp_k,
            'room_temp_k': room_temp_k,
            'implementation_loss_db': implementation_loss_db,
            }



#!/usr/bin/python

import math

from .. import utils


def _mean_orbit_altitude_km(model):
    return (model.apoapsis_altitude_km + model.periapsis_altitude_km) / 2


def _geo_altitude_km(model):
    return model.geo_radius_km - model.earth_radius_km


def _tx_distance_to_geo_km(model):
    # FIXME: rename to show it is the smallest value
    distance = model.geo_altitude_km
    if model.is_downlink:
        distance -= model.apoapsis_altitude_km
    return distance


def _slant_range_km(model):
    R = model.earth_radius_km
    h = model.mean_orbit_altitude_km + R
    e = math.radians(model.min_elevation_deg)
    return R * ((((h**2/R**2)-(math.cos(e))**2)**0.5) - math.sin(e))


def _periapsis_slant_range_km(model):    
    R = model.earth_radius_km
    h = model.periapsis_altitude_km + R
    e = math.radians(model.min_elevation_deg)
    return R * ((((h**2/R**2)-(math.cos(e))**2)**0.5) - math.sin(e))


def _satellite_antenna_angle_deg(model):
    R = model.earth_radius_km
    h = model.mean_orbit_altitude_km + R
    e = math.radians(model.min_elevation_deg)
    r = model.slant_range_km
    tmp = min(1.0, ((r**2+h**2-R**2)/(2*r*h)))
    return math.degrees(math.acos(tmp))


class Geometry(object):

    def __init__(self,
                 apoapsis_altitude_km=650,
                 periapsis_altitude_km=650,
                 min_elevation_deg=10,
                 earth_radius_km=6378.14,
                 geo_radius_km=42164):

        self.tribute = {
            # calculators
            'slant_range_km': _slant_range_km,
            'satellite_antenna_angle_deg': _satellite_antenna_angle_deg,
            'mean_orbit_altitude_km': _mean_orbit_altitude_km,
            'geo_altitude_km': _geo_altitude_km,
            'tx_distance_to_geo_km': _tx_distance_to_geo_km,
            'periapsis_slant_range_km': _periapsis_slant_range_km,

            # constants
            'geo_radius_km': geo_radius_km,
            'apoapsis_altitude_km': apoapsis_altitude_km,
            'periapsis_altitude_km': periapsis_altitude_km,
            'earth_radius_km': earth_radius_km,
            'min_elevation_deg': min_elevation_deg,
            }



#!/usr/bin/python

import math

from .. import utils


class Code(object):

    def __init__(self, name, tx_eff, rx_eff, esn0_db):
        self.name = name
        self.tx_eff = tx_eff
        self.rx_eff = rx_eff
        self.esn0_db = esn0_db
        self.ebn0_db = esn0_db - utils.to_db(rx_eff)


# http://www.etsi.org/deliver/etsi_en/302300_302399/30230702/01.01.01_20
#       /en_30230702v010101a.pdf
# Page 52
# One notable exception is that the transmit spectral efficiency
# defaults to 0.8 * the receive spectral efficiency to leave room for
# rolloff.  Feel free to override
NORMAL_DVBS2X_PERFORMANCE = [
    Code("QPSK 13/45", 0.454244, 0.567805, -2.030000),
    Code("QPSK 9/20", 0.711308, 0.889135, 0.220000),
    Code("QPSK 11/20", 0.870865, 1.088581, 1.450000),
    Code("8APSK 5/9-L", 1.317769, 1.647211, 4.730000),
    Code("8APSK 26/45-L", 1.370881, 1.713601, 5.130000),
    Code("8PSK 23/36", 1.516938, 1.896173, 6.120000),
    Code("8PSK 25/36", 1.649718, 2.062148, 7.020000),
    Code("8PSK 13/18", 1.716109, 2.145136, 7.490000),
    Code("16APSK 1/2-L", 1.577802, 1.972253, 5.970000),
    Code("16APSK 8/15-L", 1.683880, 2.104850, 6.550000),
    Code("16APSK 5/9-L", 1.754598, 2.193247, 6.840000),
    Code("16APSK 26/45", 1.825316, 2.281645, 7.510000),
    Code("16APSK 3/5", 1.896034, 2.370043, 7.800000),
    Code("16APSK 3/5-L", 1.896034, 2.370043, 7.410000),
    Code("16APSK 28/45", 1.966753, 2.458441, 8.100000),
    Code("16APSK 23/36", 2.019791, 2.524739, 8.380000),
    Code("16APSK 2/3-L", 2.108189, 2.635236, 8.430000),
    Code("16APSK 25/36", 2.196587, 2.745734, 9.270000),
    Code("16APSK 13/18", 2.284985, 2.856231, 9.710000),
    Code("16APSK 7/9", 2.461780, 3.077225, 10.650000),
    Code("16APSK 77/90", 2.709294, 3.386618, 11.990000),
    Code("32APSK 2/3-L", 2.631602, 3.289502, 11.100000),
    Code("32APSK 32/45", 2.808154, 3.510192, 11.750000),
    Code("32APSK 11/15", 2.896429, 3.620536, 12.170000),
    Code("32APSK 7/9", 3.072981, 3.841226, 13.050000),
    Code("64APSK 32/45-L", 3.365142, 4.206428, 13.980000),
    Code("64APSK 11/15", 3.470927, 4.338659, 14.810000),
    Code("64APSK 7/9", 3.682498, 4.603122, 15.470000),
    Code("64APSK 4/5", 3.788283, 4.735354, 15.870000),
    Code("64APSK 5/6", 3.946961, 4.933701, 16.550000),
    Code("128APSK 3/4", 4.130598, 5.163248, 17.730000),
    Code("128APSK 7/9", 4.284445, 5.355556, 18.530000),
    Code("256APSK 29/45-L", 4.052552, 5.065690, 16.980000),
    Code("256APSK 2/3-L", 4.193211, 5.241514, 17.240000),
    Code("256APSK 31/45-L", 4.333870, 5.417338, 18.100000),
    Code("256APSK 32/45", 4.474530, 5.593162, 18.590000),
    Code("256APSK 11/15-L", 4.615190, 5.768987, 18.840000),
    Code("256APSK 3/4", 4.720684, 5.900855, 19.570000),
    ]

PERFECT_DVBS2X_PERFORMANCE = [ Code(c.name, c.rx_eff, c.rx_eff, c.esn0_db)
                               for c in NORMAL_DVBS2X_PERFORMANCE ]


def _modulation_code_lookup_table(model):
    e = model.enum

    allocation = model.allocation_hz

    best_option = None
    best_bitrate = -1

    retval = {}

    for code in model.modulation_performance_table:
        retval[code.ebn0_db] = code

    return retval


def _max_allowable_bitrate_hz(model):
    return model.best_modulation_code.tx_eff * model.allocation_hz


def __max_bitrate_hz(model, code, additional_rx_losses_db):
    R_db_hz = (model.cn0_db
               - additional_rx_losses_db
               - model.target_margin_db
               - code.ebn0_db)

    max_R = model.allocation_hz * code.tx_eff

    return min(utils.from_db(R_db_hz), max_R)


def _best_modulation_code(model):
    e = model.enum

    if 1 == len(model.modulation_performance_table):
        return model.modulation_performance_table[0]

    def __rate_for(code):
        # DANGER WILL ROBINSON!!
        model.override(e.best_modulation_code, code)
        added_loss = model.cached_calculate(e.additional_rx_losses_db,
                                            clear_stack=True)
        model.revert(e.best_modulation_code)
        # DANGER WILL ROBINSON!!
        return __max_bitrate_hz(model, code, added_loss)

    # FIXME: Use an O(log(n)) algorithm here...for kicks and cycles
    prev_R = 0
    retval = None
    for code in model.modulation_performance_table:
        R = __rate_for(code)
        if R > prev_R:
            prev_R = R
            retval = code

    return retval


def _rx_spectral_efficiency_bps_per_hz(model):
    return model.best_modulation_code.rx_eff


def _tx_spectral_efficiency_bps_per_hz(model):
    return model.best_modulation_code.tx_eff


def _required_demod_ebn0_db(model):
    return model.best_modulation_code.ebn0_db


def _max_bitrate_hz(model):
    return __max_bitrate_hz(model,
                            model.best_modulation_code,
                            model.additional_rx_losses_db)


class Modulation(object):
    """Modulation tributary

    This is designed to allow for multiple modulation code options
    (such as found in DVB-S2X.  It includes the logic to solve for the
    most appropriate code option under the circumstances, as well.
    """

    def __init__(self, name='DVB-S2X', perf=None):
        """Create a new modulation tributary

        perf=[Code(), Code(), ...]

        If you don't want to use DVB-S2X, override the performance
        table when creating the modulation object.
        """

        if perf is None:
            perf = NORMAL_DVBS2X_PERFORMANCE

        self.tribute = {
            # calculators
            'max_allowable_bitrate_hz': _max_allowable_bitrate_hz,
            'required_demod_ebn0_db': _required_demod_ebn0_db,
            'best_modulation_code': _best_modulation_code,
            'tx_spectral_efficiency_bps_per_hz': _tx_spectral_efficiency_bps_per_hz,
            'rx_spectral_efficiency_bps_per_hz': _rx_spectral_efficiency_bps_per_hz,
            'max_bitrate_hz': _max_bitrate_hz,
            'modulation_code_lookup_table': _modulation_code_lookup_table,

            # constants
            'modulation_name': name,
            'modulation_performance_table': perf,
            }



#!/usr/bin/python

from setuptools import setup
import os

data_dirs = {'tex': '/pylink/tex'}

cur_dir = os.path.dirname(os.path.realpath(__file__))

data_files = []

with open(os.path.join(cur_dir, 'README.md'), 'r') as fh:
    long_description = fh.read()

for srcdir, dst in data_dirs.items():
    srcpath = os.path.join(cur_dir, srcdir)
    files = [os.path.join(srcdir, f) for f in os.listdir(srcpath)]
    data_files.append((dst, files,))

setup(
    name='pylink-satcom',
    version='0.6',
    author='Harrison Caudill',
    author_email='harrison@hypersphere.org',
    description='Python Link Budget System',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/harrison-caudill/pylink',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Ham Radio',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Topic :: Scientific/Engineering :: Physics',
        ],
    data_files=data_files,
    install_requires=[
        'matplotlib',
        'jinja2',
        'scipy',
        'numpy',
        ],
    package_dir={
        'tributaries':'tributaries',
        },
    packages=[
        'pylink',
        'pylink.tributaries'
        ]
    )



#!/usr/bin/python

import math

from .. import utils


class Transmitter(object):

    def __init__(self,
                 tx_power_at_pa_dbw=3,
                 rf_chain=[],
                 name='Transmitter'):

        self.tribute = {
            # calculators

            # constants
            'transmitter_rf_chain': rf_chain,
            'transmitter_name': name,
            'tx_power_at_pa_dbw': tx_power_at_pa_dbw,
            }



#!/usr/bin/python


class Element(object):
    """RF Chain Element container object
    """

    def __init__(self, gain_db, noise_figure_db, name, **kwargs):
        """New RF Chain container

        Add any kwargs for any desired metadata for this object.  For
        example, the link to the datasheet, test report, or maybe its
        internal and/or external part numbers.
        """
        self.gain_db = float(gain_db)
        self.noise_figure_db = float(noise_figure_db)
        self.name = name
        self.meta = kwargs
